#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

/*#define DEBUG*/

#include "realfftf.h"

void main(int argc,char *argv[])
{
	int i;                  /* Generic index */
	FILE *fp1,*fp2;         /* Input and output file pointers */
	char filename[20];      /* Used for filename inputs */
	int points;             /* Number of samples per FFT (power of 2) */
	fft_type *data;         /* Pointer to the data passed to the FFT */
	fft_type *window;		   /* Pointer to the windowing function */
	fft_type *output;       /* Pointer to the output buffer */
	fft_type re,im;		   /* Variables used for computing magnitude data */
	unsigned char *ptr_uc;  /* Pointer to 8-bit unsigned inputs */
	short *ptr_short;       /* Pointer to 16-bit signed inputs */
	char buf[10];           /* General storage buffer */
	int read;				   /* Number of data points read from the file */
	long length;            /* Length of the input file */
	int badfile;            /* Flag indicating a bad file header */

	/* Header for a .WAV file */
	struct WAVHDR
	{
		char riff[4];        /* Must be "RIFF" */
		long length;         /* Total length of file */
		char wavefmt[8];     /* Must be "WAVEfmt " */
		long hdr_len;        /* length of header */
		int style;           /* Style: 1=unsigned */
		int channels;        /* channels: mono or stereo */
		long rate;           /* samples per second */
		long bytesps;        /* average bytes per second */
		int align;           /* bytes per clock tick */
		int size;            /* bits per sample */
	} wavhdr;

	if(argc>1)
		fp1=fopen(argv[1],"rb");
	else
	{
		printf("Input file name: ");
		scanf("%s",filename);
		fp1=fopen(filename,"rb");
	}

	if(!fp1)
	{
		puts("Error opening input file!");
		exit(1);
	}

	if(argc>2)
		fp2=fopen(argv[2],"wb");
	else
	{
		printf("Output file name: ");
		scanf("%s",filename);
		fp2=fopen(filename,"wb");
	}

	if(!fp2)
	{
		puts("Error opening output file!");
		exit(1);
	}

	if(argc>3)
		points=atoi(argv[3]);
	else
	{
		printf("Number of points: ");
		scanf("%d",&points);
	}

	/* Convert number of points to a power of 2 */
	for(i=0;points>1;points>>=1,i++);
	points<<=i;
	data=(fft_type *)malloc(points*sizeof(fft_type));
	window=(fft_type *)malloc(points*sizeof(fft_type));
	output=(fft_type *)malloc(points*sizeof(fft_type));
	if((window==NULL)||(data==NULL)||(output==NULL))
	{
		puts("Error allocating memory");
		exit(1);
	}

	/* Create the window function (Blackman) */
	for(i=0;i<points;i++)
		window[i]=0.42-0.5*cos(2*M_PI*i/points)+0.08*cos(4*M_PI*i/points);

	/* Read the .WAV header, and verify authenticity */
	fread(&wavhdr,sizeof(struct WAVHDR),1,fp1);

	#ifdef DEBUG
		printf("riff: %.4s\n",wavhdr.riff);
		printf("length: %ld\n",wavhdr.length);
		printf("wavefmt: %.8s\n",wavhdr.wavefmt);
		printf("header length: %ld\n",wavhdr.hdr_len);
		printf("style: %d\n",wavhdr.style);
		printf("channels: %d\n",wavhdr.channels);
		printf("rate: %ld\n",wavhdr.rate);
		printf("bytes per second: %ld\n",wavhdr.bytesps);
		printf("align: %d\n",wavhdr.align);
		printf("size: %d\n",wavhdr.size);
	#endif

	/* Do some error checking */
	badfile=0;
	if(strncmp(wavhdr.riff,"RIFF",4)!=0) badfile=1;
	if(strncmp(wavhdr.wavefmt,"WAVEfmt ",8)!=0) badfile=1;
	if(badfile)
	{
		puts("Input file does not seem to be a .wav file");
		exit(1);
	}

	/* Seek to the end of the header */
	fseek(fp1,wavhdr.hdr_len-16,SEEK_CUR);
	i=0;
	while((++i<10) && !feof(fp1))
	{
		fread(buf,4,1,fp1);
		fread(&length,sizeof(long),1,fp1);
		#ifdef DEBUG
			printf("%4.4s, %ld\n",buf,length);
		#endif
		if(strncmp(buf,"data",4)==0)
			break;
		else
			fseek(fp1,length,SEEK_CUR);  /* Skip this block */
	}
	if((i>=10) || feof(fp1))
	{
		puts("This file really confuses me!");
		exit(1);
	}
	if(length<0)
	{
		puts("This file claims it has no samples.");
		exit(1);
	}
	if(wavhdr.channels!=1)
	{
		puts("Sorry, I'm too stupid to process stereo files.");
		exit(1);
	}
	if((wavhdr.size!=8) && (wavhdr.size!=16))
	{
		printf("The header says you have %d bits per sample.  I'm confused.\n",wavhdr.size);
		exit(1);
	}

	length=length*wavhdr.size/8;
	printf("Performing %d point FFTs on a %d-bit wave file with %ld samples",
			 points,wavhdr.size,length);

	InitializeFFT(points);

	if(wavhdr.size==8)
	{
		ptr_uc=(unsigned char *)malloc(points*sizeof(unsigned char));
		if(ptr_uc==NULL)
		{
			puts("Error allocating memory");
			exit(1);
		}
	}
	else
	{
		ptr_short=(short *)malloc(points*sizeof(short));
		if(ptr_short==NULL)
		{
			puts("Error allocating memory");
			exit(1);
		}
	}
	while(!feof(fp1))
	{
		if(wavhdr.size==8)
		{
			/* Processing for 8-bit unsigned data */
			read=fread(ptr_uc,sizeof(unsigned char),points,fp1);
			/* Window the data */
			for(i=0;i<points;i++)
				data[i]=((fft_type)(ptr_uc[i])-128.0)*window[i];
		}
		else
		{
			/* Processing for 16-bit signed data */
			read=fread(ptr_short,sizeof(short),points,fp1);
			/* Window the data */
			for(i=0;i<points;i++)
				data[i]=(fft_type)(ptr_short[i])*window[i];
		}
		if(read==points)
		{
			/* Perform an FFT */
			RealFFT(data);
			/* Compute the magnitude information */
			for(i=0;i<(points/2);i++)
			{
				re=data[BitReversed[i]];
				im=data[BitReversed[i]+1];
				output[i]=sqrt(re*re+im*im);
			}
			/* Save the output */
			fwrite(output,sizeof(fft_type),points/2,fp2);
		}
	}

	fclose(fp1);
	fclose(fp2);

	/* Free up the FFT buffers */
	EndFFT();
}
